@@ -1,7 +1,7 @@
 # coding=utf-8
 """A Python Package for parallelizing ssh connections via multi-threading and multi-processing.
 """
-# Copyright (C) 2015 Jesse Almanrode
+# Copyright (C) 2015-2017 Jesse Almanrode
 #
 #     This program is free software: you can redistribute it and/or modify
 #     it under the terms of the GNU Lesser General Public License as published by
@@ -20,7 +20,8 @@ from pkg_resources import get_distribution, DistributionNotFound
 
 # For backwards compatibility
 from sshreader.ssh import SSH, shell_command, do_shell_script, envvars
-from sshreader.sshreader import ServerJob, Hook, sshread, print_results, echo, cpusoftlimit, cpuhardlimit
+from sshreader.sshreader import ServerJob, Hook, sshread, print_results, echo, cpusoftlimit, cpuhardlimit, threadlimit
+import logging
 
 __author__ = 'Jesse Almanrode (jesse@almanrode.com)'
 try:
@@ -28,3 +29,6 @@ try:
 except DistributionNotFound:
     __version__ = 'UNKNOWN'
 __all__ = ['sshreader', 'ssh']
+
+logging.basicConfig()
+logging.getLogger(__name__).addHandler(logging.NullHandler())
