@@ -3,7 +3,7 @@
 """A wrapper for Paramiko that attempts to make ssh sessions easier to work with.  It also contains the
 shell_command function for running local shell scripts!
 """
-# Copyright (C) 2015 Jesse Almanrode
+# Copyright (C) 2015-2017 Jesse Almanrode
 #
 #     This program is free software: you can redistribute it and/or modify
 #     it under the terms of the GNU Lesser General Public License as published by
@@ -18,41 +18,41 @@ shell_command function for running local shell scripts!
 #     You should have received a copy of the GNU Lesser General Public License
 #     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 from __future__ import print_function
+from collections import namedtuple, OrderedDict
+from getpass import getuser
+from subprocess import Popen, PIPE, STDOUT
 import logging
 import os
 import paramiko
+import socket
 import warnings
-from collections import namedtuple
-from getpass import getuser
-from subprocess import Popen, PIPE, STDOUT
 
 __author__ = 'Jesse Almanrode (jesse@almanrode.com)'
 
-# Using namedtuple because... why not?
 ShellCommand = namedtuple('ShellCommand', ['cmd', 'stdout', 'stderr', 'return_code'])
 ShellCommandCombined = namedtuple('ShellCommandCombined', ['cmd', 'stdout', 'return_code'])
-EnvVars = namedtuple('EnvVars', ['username', 'rsa_key', 'dsa_key'])
 
 
 def envvars():
     """ Attempt to determine the current username and location of any ssh keys.  If any value is unable to be determined
     it is returned as 'None'.
 
-    :return: NamedTuple of (username, rsa_key, dsa_key)
+    :return: NamedTuple of (username, rsa_key, dsa_key, ecdsa_key)
     """
-    user = None
-    rsa_key = None
-    dsa_key = None
+    env = OrderedDict(user=None, rsa_key=None, dsa_key=None, ecdsa_key=None)
+    EnvVars = namedtuple('EnvVars', env.keys())
     if os.getlogin() == getuser():
-        user = getuser()
+        env['user'] = getuser()
     userhome = os.path.expanduser('~')
     if os.path.exists(userhome + "/.ssh"):
         keyfiles = os.listdir(userhome + "/.ssh")
         if "id_rsa" in keyfiles:
-            rsa_key = userhome + "/.ssh/id_rsa"
+            env['rsa_key'] = userhome + "/.ssh/id_rsa"
         if "id_dsa" in keyfiles:
-            dsa_key = userhome + "/.ssh/id_dsa"
-    return EnvVars(user, rsa_key, dsa_key)
+            env['dsa_key'] = userhome + "/.ssh/id_dsa"
+        if 'id_ecdsa' in keyfiles:
+            env['ecdsa_key'] = userhome + '/.ssh/id_ecdsa'
+    return EnvVars(**env)
 
 
 def shell_command(command, combine=False, decodebytes=True):
@@ -89,7 +89,7 @@ def do_shell_script(command, combine=False):
 
     .. warning::
 
-        This call will be deprecated in v4.0
+        This call will be removed in v4.0
 
     :param command: String containing the shell script to run
     :param combine: Combine stderr and stdout in output
@@ -124,8 +124,8 @@ class SSH(object):
             self.keyfile = keyfile
         self.port = port
         self.timeout = timeout
-        self.connection = paramiko.SSHClient()
-        self.connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())
+        self._connection = paramiko.SSHClient()
+        self._connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())
         if connect:
             self.__connect()
 
@@ -136,7 +136,7 @@ class SSH(object):
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.__close()
-        
+
     def sftp_put(self, srcfile, dstfile):
         """ Use the SFTP subsystem of OpenSSH to copy a local file to a remote host
 
@@ -144,9 +144,14 @@ class SSH(object):
         :param dstfile: Path to the remote file
         :return: Result of paramiko.SFTPClient.put()
         """
-        sftp = paramiko.SFTPClient.from_transport(self.connection.get_transport())
-        result = sftp.put(os.path.expanduser(srcfile), os.path.expanduser(dstfile), confirm=True)
-        sftp.close()
+        if self.__is_alive() is False:
+            raise paramiko.SSHException("Connection is not established")
+        sftp = paramiko.SFTPClient.from_transport(self._connection.get_transport())
+        try:
+            result = sftp.put(os.path.expanduser(srcfile), os.path.expanduser(dstfile), confirm=True)
+        except IOError as err:
+            sftp.close()
+            raise err
         return result
 
     def sftp_get(self, srcfile, dstfile):
@@ -156,8 +161,14 @@ class SSH(object):
         :param dstfile: Path to the local file
         :return: Result of paramiko.SFTPClient.get()
         """
-        sftp = paramiko.SFTPClient.from_transport(self.connection.get_transport())
-        result = sftp.get(os.path.expanduser(srcfile), os.path.expanduser(dstfile))
+        if self.__is_alive() is False:
+            raise paramiko.SSHException("Connection is not established")
+        sftp = paramiko.SFTPClient.from_transport(self._connection.get_transport())
+        try:
+            result = sftp.get(os.path.expanduser(srcfile), os.path.expanduser(dstfile))
+        except IOError as err:
+            sftp.close()
+            raise err
         sftp.close()
         return result
 
@@ -166,7 +177,7 @@ class SSH(object):
 
         :param command: The command to run
         :param timeout: Timeout for the command
-        :param combine: Combine stderr and stdout
+        :param combine: Combine stderr and stdout (pseudo TTY)
         :param decodebytes: Decode bytes objects to unicode strings
         :return: Namedtuple of (cmd, stdout, stderr, return_code) or (cmd, stdout, return_code)
         :raises: SSHException
@@ -174,28 +185,28 @@ class SSH(object):
         if self.__is_alive() is False:
             raise paramiko.SSHException("Connection is not established")
         if combine:
-            # http://stackoverflow.com/questions/3823862/paramiko-combine-stdout-and-stderr
-            chan = self.connection.get_transport().open_session()
-            chan.settimeout(timeout)
-            chan.get_pty()
-            stdout_file = chan.makefile()
-            chan.exec_command(command)
-            stdout = stdout_file.read()
-            stdout_file.close()
-            if decodebytes:
-                result = ShellCommandCombined(cmd=command, stdout=stdout.decode().strip(),
-                                              return_code=chan.recv_exit_status())
-            else:
-                result = ShellCommandCombined(cmd=command, stdout=stdout.strip(), return_code=chan.recv_exit_status())
+            try:
+                stdin, stdout, stderr = self._connection.exec_command(command, timeout=timeout, get_pty=True)
+                if decodebytes:
+                    result = ShellCommandCombined(cmd=command, stdout=stdout.read().decode().strip(),
+                                                  return_code=stdout.channel.recv_exit_status())
+                else:
+                    result = ShellCommandCombined(cmd=command, stdout=stdout.read().strip(),
+                                                  return_code=stdout.channel.recv_exit_status())
+            except (paramiko.buffered_pipe.PipeTimeout, socket.timeout):
+                result = ShellCommandCombined(cmd=command, stdout='Command timed out', return_code=124)
         else:
-            stdin, stdout, stderr = self.connection.exec_command(command, timeout=timeout)
-            if decodebytes:
-                result = ShellCommand(cmd=command, stdout=stdout.read().decode().strip(),
-                                      stderr=stderr.read().decode().strip(),
-                                      return_code=stdout.channel.recv_exit_status())
-            else:
-                result = ShellCommand(cmd=command, stdout=stdout.read().strip(), stderr=stderr.read().strip(),
-                                      return_code=stdout.channel.recv_exit_status())
+            try:
+                stdin, stdout, stderr = self._connection.exec_command(command, timeout=timeout)
+                if decodebytes:
+                    result = ShellCommand(cmd=command, stdout=stdout.read().decode().strip(),
+                                          stderr=stderr.read().decode().strip(),
+                                          return_code=stdout.channel.recv_exit_status())
+                else:
+                    result = ShellCommand(cmd=command, stdout=stdout.read().strip(), stderr=stderr.read().strip(),
+                                          return_code=stdout.channel.recv_exit_status())
+            except (paramiko.buffered_pipe.PipeTimeout, socket.timeout):
+                result = ShellCommand(cmd=command, stdout='', stderr='Command timed out', return_code=124)
         return result
 
     def close(self):
@@ -203,7 +214,7 @@ class SSH(object):
 
         :return: None
         """
-        self.connection.close()
+        self._connection.close()
         return None
 
     def is_alive(self):
@@ -212,10 +223,10 @@ class SSH(object):
         :return: True or False
         :raises: SSHException
         """
-        if self.connection.get_transport() is None:
+        if self._connection.get_transport() is None:
             return False
         else:
-            if self.connection.get_transport().is_alive():
+            if self._connection.get_transport().is_alive():
                 return True
             else:
                 raise paramiko.SSHException("Unable to determine state of ssh session")
@@ -231,22 +242,22 @@ class SSH(object):
         :return: True
         :raises: SSHException
         """
-        logging.basicConfig()  # http://stackoverflow.com/questions/26659772/
+        paramiko.util.logging.getLogger().setLevel(logging.CRITICAL)  # Keeping paramiko from logging errors to stdout
         if self.__is_alive():
             raise paramiko.SSHException("Connection is already established")
         if self.keyfile is not None:
             if self.username is not None:  # Key file with a custom username!
-                self.connection.connect(self.host, port=self.port, username=self.username,
-                                        key_filename=self.keyfile, timeout=self.timeout, look_for_keys=False)
+                self._connection.connect(self.host, port=self.port, username=self.username,
+                                         key_filename=self.keyfile, timeout=self.timeout, look_for_keys=False)
             else:
-                self.connection.connect(self.host, port=self.port, key_filename=self.keyfile,
-                                        timeout=self.timeout, look_for_keys=False)
+                self._connection.connect(self.host, port=self.port, key_filename=self.keyfile,
+                                         timeout=self.timeout, look_for_keys=False)
         else:  # Username and password combo
             if self.username is None or self.password is None:
                 raise paramiko.SSHException("You must provide a username and password or supply an SSH key")
             else:
-                self.connection.connect(self.host, port=self.port, username=self.username,
-                                        password=self.password, timeout=self.timeout, look_for_keys=False)
+                self._connection.connect(self.host, port=self.port, username=self.username,
+                                         password=self.password, timeout=self.timeout, look_for_keys=False)
         return True
 
     # Privatizing some of the functions so SSH can be subclassed
